import vlsir
from vlsirtools.netlist import (
    netlist_from_proto,
)


def test_netlist_from_proto1():
    """Netlisting from protobuf, smoke test"""

    netlist_path = "scratch/test_netlist1.sp"
    netlist_from_proto(
        vlsir.netlist.NetlistInput(
            pkg=vlsir.circuit.Package(),
            netlist_path=netlist_path,
            fmt=vlsir.netlist.NetlistFormat.SPICE,
            result_path="scratch/test_netlist1.result.pb",  # (this doesn't do anything)
        )
    )
    readback = open(netlist_path, "r").read()
    assert len(readback) > 0
    # These are tests we often come to regret, checking netlist literal content. But gotta do something for now!
    assert "* Anonymous `circuit.Package`" in readback
    assert "* Generated by `vlsirtools.SpiceNetlister`" in readback



""" 
# Spice-Class Simulator Interface 
"""

# Std-Lib Imports
import os, subprocess
import concurrent.futures
from typing import Union, Optional, Sequence, TypeVar
from enum import Enum
from pathlib import Path
from textwrap import dedent
from dataclasses import dataclass, field, replace

# Local/ Project Dependencies
import vlsir.spice_pb2 as vsp
from vlsir.spice_pb2 import *  # Not used here intentionally, but "re-exported"
from . import sim_data as sd


class ResultFormat(Enum):
    """Enumerated Result Formats"""

    SIM_DATA = "sim_data"  # Types defined in the `sim_data` module
    VLSIR_PROTO = "vlsir_proto"  # `vsp.SimResults` and related protobuf-defined types


# Union of the two result-types, returned by many simulation methods
SimResultUnion = Union[vsp.SimResult, sd.SimResult]


class SupportedSimulators(Enum):
    """Enumerated, Internally-Defined Spice-Class Simulators"""

    SPECTRE = "spectre"
    XYCE = "xyce"
    NGSPICE = "ngspice"


def default() -> Optional[SupportedSimulators]:
    """Get the default simulator, for this Python-process and its environment.
    This largely consists of a priority-ordered walk through `SupportedSimulators`,
    returning the first whose `available()` method indicates its availability.
    Returns `None` if no such simulator appears available."""

    from .spectre import available as spectre_available
    from .xyce import available as xyce_available
    from .ngspice import available as ngspice_available

    if spectre_available():
        return SupportedSimulators.SPECTRE
    if xyce_available():
        return SupportedSimulators.XYCE
    if ngspice_available():
        return SupportedSimulators.NGSPICE
    return None  # Nothing found


@dataclass
class SimOptions:
    """Options to `vsp.Sim` which are *not* passed along to the simulator.
    I.e. options which effect the Python-process invoking simulation,
    but not the internals of the simulation itself."""

    # Simulator. FIXME: debatable whether to include this.
    simulator: SupportedSimulators = field(default_factory=default)

    # In-memory results format
    fmt: ResultFormat = ResultFormat.VLSIR_PROTO

    # Simulation run-directory. Uses a `tempdir` if unspecified.
    rundir: Optional[os.PathLike] = None


# Shorthand type alias for "an element or list thereof", used by all the call signatures below
T = TypeVar("T")
OneOrMore = Union[T, Sequence[T]]


@dataclass
class SimInputAndOptions:
    inp: vsp.SimInput
    opts: SimOptions


def sim(
    inp: OneOrMore[vsp.SimInput], opts: Optional[SimOptions] = None
) -> OneOrMore[SimResultUnion]:
    """
    Concurrently execute one or more `vlir.spice.Sim`.
    Dispatches across `SupportedSimulators` specified in `SimOptions` `opts`.
    Uses the default `Simulator` as detected by the `default` method if no `simulator` is specified.
    """
    from .xyce import XyceSim
    from .spectre import SpectreSim
    from .ngspice import NGSpiceSim

    if opts is None:  # Create the default `SimOptions`
        opts = SimOptions()

    if opts.simulator is None:  # If we didn't specify or find a simulator, fail.
        msg = f"vlsirtools.spice: No Supported Simulators available for call to `sim()`"
        raise RuntimeError(msg)

    # Get the per-simulator callable
    if opts.simulator == SupportedSimulators.XYCE:
        cls = XyceSim
    elif opts.simulator == SupportedSimulators.SPECTRE:
        cls = SpectreSim
    elif opts.simulator == SupportedSimulators.NGSPICE:
        cls = NGSpiceSim
    else:
        raise ValueError(f"Unsupported simulator: {opts.simulator}")

    # Sort out the difference between "One" "OrMore" cases of input
    # For a single `SimInput`, create a list, but note we only want to return a single `SimResult`
    inp_is_a_single_sim = False
    if not isinstance(inp, Sequence):
        inp = [inp]
        inp_is_a_single_sim = True

    inputs_and_options: List[SimInputAndOptions] = []
    for idx, x in enumerate(inp):
        if not isinstance(x, vsp.SimInput):
            raise TypeError(f"Expected `vsp.SimInput`, got {x}")
        if len(inp) > 1 and opts.rundir is not None:
            # Create sub-directories of the form `/run/dir/0`, `/run/dir/1`, etc.
            this_opts = replace(opts, rundir=Path(opts.rundir) / str(idx))
            io = SimInputAndOptions(inp=x, opts=this_opts)
        else:
            io = SimInputAndOptions(inp=x, opts=opts)
        inputs_and_options.append(io)

    # And do the real work, invoking the target simulator
    # Note the list of `SimResult`s is ordered per the order of `SimInput`s.
    with concurrent.futures.ThreadPoolExecutor() as executor:
        results = list(executor.map(cls.apply, inputs_and_options))

    # For the sequence of inputs case, return the sequence of results that came back
    if not inp_is_a_single_sim:
        return results

    # Unpack the single-input case
    if len(results) != 1:
        raise RuntimeError("Expected a single result")
    return results[0]


class SimError(Exception):
    """Exception raised when a simulation fails."""

    def __init__(self, sim: "Sim", stdout: bytes, stderr: bytes) -> None:
        s = dedent(
            f"""
            rundir: {str(sim.rundir)}
            stdout: {str(stdout)}
            stderr: {str(stderr)}
        """
        )
        super().__init__(s)
        self.rundir = sim.rundir
        self.stdout = stdout
        self.stderr = stderr


class SimProcessError(SimError):
    """Exception raised when an external simulator process fails."""

    def __init__(self, sim: "Sim", e: subprocess.CalledProcessError) -> None:
        super().__init__(sim=sim, stdout=e.stdout, stderr=e.stderr)



""" 
# Netlist Formats
"""

# Std-Lib Imports
from enum import Enum
from typing import Union

# Local Imports
import vlsir

# Locally-defined netlister classes
from .spectre import SpectreNetlister
from .verilog import VerilogNetlister
from .spice import (
    SpiceNetlister,
    HspiceNetlister,
    CdlNetlister,
    XyceNetlister,
    NgspiceNetlister,
)


class NetlistFormat(Enum):
    """Enumeration of available formats.
    Includes string-value conversion."""

    VERILOG = "verilog"
    SPECTRE = "spectre"

    # Spice Dialects
    SPICE = "spice"
    NGSPICE = "ngspice"
    XYCE = "xyce"
    HSPICE = "hspice"
    CDL = "cdl"

    @staticmethod
    def get(spec: "NetlistFormatSpec") -> "NetlistFormat":
        """Get the format specified by `spec`, in either enum or string terms.
        Only does real work in the case when `spec` is a string, otherwise returns it unchanged.
        """
        if isinstance(spec, (NetlistFormat, str)):
            return NetlistFormat(spec)
        raise TypeError

    def netlister(self) -> type:
        """Get the paired netlister-class"""
        if self == NetlistFormat.SPECTRE:
            return SpectreNetlister
        if self == NetlistFormat.VERILOG:
            return VerilogNetlister
        if self == NetlistFormat.SPICE:
            return SpiceNetlister
        if self == NetlistFormat.HSPICE:
            return HspiceNetlister
        if self == NetlistFormat.NGSPICE:
            return NgspiceNetlister
        if self == NetlistFormat.XYCE:
            return XyceNetlister
        if self == NetlistFormat.CDL:
            return CdlNetlister
        raise ValueError(f"Unknown NetlistFormat: {self}")

    def to_proto(self) -> vlsir.netlist.NetlistFormat:
        """Convert a `NetlistFormat` to a protobuf `NetlistFormat`."""
        F = vlsir.netlist.NetlistFormat
        if self == NetlistFormat.SPECTRE:
            return F.SPECTRE
        if self == NetlistFormat.SPICE:
            return F.SPICE
        if self == NetlistFormat.NGSPICE:
            return F.NGSPICE
        if self == NetlistFormat.XYCE:
            return F.XYCE
        if self == NetlistFormat.HSPICE:
            return F.HSPICE
        if self == NetlistFormat.CDL:
            return F.CDL
        if self == NetlistFormat.VERILOG:
            return F.VERILOG
        raise ValueError(f"Unknown NetlistFormat: {self}")

    @staticmethod
    def from_proto(proto: vlsir.netlist.NetlistFormat) -> "NetlistFormat":
        """Convert a protobuf `NetlistFormat` to a `NetlistFormat` enum."""
        F = vlsir.netlist.NetlistFormat
        if proto == F.UNSPECIFIED or proto == F.SPECTRE:
            return NetlistFormat.SPECTRE
        if proto == F.SPICE:
            return NetlistFormat.SPICE
        if proto == F.NGSPICE:
            return NetlistFormat.NGSPICE
        if proto == F.XYCE:
            return NetlistFormat.XYCE
        if proto == F.HSPICE:
            return NetlistFormat.HSPICE
        if proto == F.CDL:
            return NetlistFormat.CDL
        if proto == F.VERILOG:
            return NetlistFormat.VERILOG
        raise ValueError(f"Unknown NetlistFormat: {proto}")


# Type-alias for specifying format, either in enum or string terms
NetlistFormatSpec = Union[NetlistFormat, str]



""" 
# Vlsir Netlisting 
## The main `netlist` function(s).
"""

# Std-Lib Imports
from dataclasses import dataclass
from typing import IO, Optional

# Local Imports
import vlsir
from .fmt import NetlistFormat, NetlistFormatSpec


@dataclass
class NetlistOptions:
    """Options for netlisting."""

    indent: str = 2 * " "  # Indentation. Defaults to two spaces.
    width: int = 80  # Line-width. Defaults to 80.


## FIXME: add more `Netlistable`s
##Netlistable = Union[vlsir.circuit.Package]
Netlistable = vlsir.circuit.Package


def netlist(
    pkg: Netlistable,  ## FIXME: rename
    dest: IO,
    fmt: NetlistFormatSpec = "spectre",
    opts: Optional[NetlistOptions] = None,
) -> None:
    """Netlist proto-Package `pkg` to destination `dest`.

    Example usages:
    ```python
    h.netlist(pkg, dest=open('mynetlist.v', 'w'), fmt='verilog')
    ```
    ```python
    s = StringIO()
    h.netlist(pkg, dest=s, fmt='spectre')
    ```
    ```python
    import sys
    h.netlist(pkg, dest=sys.stdout, fmt='spice')
    ```

    Primary argument `pkg` must be a `vlsir.circuit.Package`.
    Destination `dest` may be anything that supports the `typing.IO` bundle,
    commonly including open file-handles. `StringIO` is particularly helpful
    for producing a netlist in an in-memory string.
    Format-specifier `fmt` may be any of the `NetlistFormatSpec` enumerated values
    or their string equivalents.
    """

    if opts is not None:
        raise NotImplementedError("NetlistOptions")  # FIXME!

    # If `fmt` is a string, turn it into an enum
    fmt_enum = NetlistFormat.get(fmt)

    # Get the corresponding `Netlister` class and instantiate it
    netlister_cls = fmt_enum.netlister()
    netlister = netlister_cls(dest=dest)

    # Write the netlist
    return netlister.write_package(pkg)


def netlist_from_proto(inp: vlsir.netlist.NetlistInput) -> vlsir.netlist.NetlistResult:
    """# Netlist a ProtoBuf-Dicatated `NetlistInput`"""
    try:
        netlist(
            pkg=inp.pkg,
            dest=open(inp.netlist_path, "w"),
            fmt=NetlistFormat.from_proto(inp.fmt),
            opts=None,
        )
    except Exception as e:
        return vlsir.netlist.NetlistResult(success=False, fail=str(e))
    return vlsir.netlist.NetlistResult(success=True)


# Set our exported content for star-imports
__all__ = [
    "netlist",
    "netlist_from_proto",
    "NetlistFormat",
    "NetlistFormatSpec",
    "NetlistOptions",
]



""" 
# Vlsir Netlisting 

Exports `vlsir.circuit.Package` and `vlsir.netlist.NetlistInput` to a netlist format.
"""

from .main import netlist, netlist_from_proto, NetlistOptions
from .fmt import NetlistFormat, NetlistFormatSpec
from .spectre import SpectreNetlister
from .verilog import VerilogNetlister
from .spice import (
    SpiceNetlister,
    HspiceNetlister,
    CdlNetlister,
    XyceNetlister,
    NgspiceNetlister,
)
